##
# geometry.itcl
#
# Routines to assist in manipulating screen co-ordinates
# in various ways. There are several groups of functions, to deal
# with screen geometry, points, vertices, and rectangles.
#

##########################################################################
#
# Author:  H. John Reekie
#
# Version: @(#)geometry.tcl	1.5    2/27/96
#
# Copyright (c) 1990-1996 The Regents of the University of California.
# All rights reserved.
# 
# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the
# above copyright notice and the following two paragraphs appear in all
# copies of this software.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
# THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# 
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
# CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
# 
# 						PT_COPYRIGHT_VERSION_2
# 						COPYRIGHTENDKEY
##########################################################################



##########################################################################
#
## Geometry functions
#
# These functions assist in dealing with directions specified
# either as:
#
# -- Compass directions: These are the four ``primary'' directions
#    _n_, _s_, _e_, and _w_, and the four ``secondary'' directions
#    _ne_, _nw_, _se_, and _sw_.
#
# -- Direction strings: These are _top_, _bottom_, _up_, _down_, _left_,
#    or _right_. _top_ is synonymous with _up_, and _bottom_ with
#    _down_. A secondary direction is specified as a list containing
#    two directions. The canonical form has the horizontal direction
#    first, but most of the procedures will accept either order.
#
# Most of the geometry functions accept arguments in the form
#
# --  direction [handed]: If only the first direction is
#     return the direction as a pair (_x_, _y_), such that moving
#     by that amount on the Tk canvas will move in the specified
#     direction.
#
#     If the second argument is present, treat it as an orthogonal
#     direction string, so the two direction strings together
#     specify the direction. The order of the two doesn't matter.
#     For example, the following all return (1,-1):
#
#         geomCartesian(ne)
#         geomCartesian(left up)
#         geomCartesian(top left)
#
#
# FIXME: Extend to deal with arbitrary angles
#



## geomHorizontal, geomVertical direction
#
# Test if a direction is in the horizontal or vertical plane.
# Accepts either compass or handedness directions.
#
proc geomHorizontal {direction} {
    return [lmember {e w left right} $direction]
}

proc geomVertical {direction} {
    return  [lmember {n s top bottom up down} $direction]
}


## geomCompass direction [handed]
#
# Translate one or two direction strings into a compass co-ordinate.
# If two strings are specified, their order does not matter.
#
# --  direction [handed]: If only the first direction is
#     given, return a primary compass direction; if both are given
#     return a secondary comp[ass direction. The order of the two
#     direction strings does not matter.
#
set _geometryCompass(left)   "w"
set _geometryCompass(right)  "e"
set _geometryCompass(top)    "n"
set _geometryCompass(bottom) "s"
set _geometryCompass(up)     "n"
set _geometryCompass(down)   "s"

proc geomCompass {direction args} {
    global _geometryCompass

    if { $args == "" } {
	return $_geometryCompass($direction)
    } else {
	return [geomCompassNormalize \
		[butt $_geometryCompass($direction) $_geometryCompass($args)]]
    }
}


## geomDirection compass
#
# Translate a compass direction into direction strings. The four primary
# co-ordinates return a single word, the four secondary directions
# return a pair of words.
#
# Arguments:
# --  compass: Any valid compass direction.
#
set _geometryDirection(n) "top"
set _geometryDirection(s) "bottom"
set _geometryDirection(e) "right"
set _geometryDirection(w) "left"

set _geometryDirection(nw) "left top"
set _geometryDirection(ne) "right top"
set _geometryDirection(sw) "left bottom"
set _geometryDirection(se) "right bottom"

proc geomDirection {compass} {
    global _geometryDirection
    return $_geometryDirection($compass)
}


## geomFlip direction [handed]
#
# Change a direction 180 degrees.
#
set _geometryFlipped(n) "s"
set _geometryFlipped(s) "n"
set _geometryFlipped(e) "w"
set _geometryFlipped(w) "e"

set _geometryFlipped(nw) "se"
set _geometryFlipped(ne) "sw"
set _geometryFlipped(sw) "ne"
set _geometryFlipped(se) "nw"

proc geomFlip {direction args} {
    global _geometryFlipped

    if { [geomIsCompass $direction] } {
	    return $_geometryFlipped($direction)
    } else {
	return [eval geomDirection \
		$_geometryFlipped([eval geomCompass $direction $args])]
    }
}


## geomIsCompass compass
#
# Test if a direction is a compass direction.
#
proc geomIsCompass {compass} {
    return [lmember {n s e w ne nw se sw} $compass]
}


## geomRotate [-anti] direction [handed]
#
# Change a direction 90 degrees. By default rotates
# clockwise, -anti option rotates counter-cw.
#
set _geometryRotated90(n) "e"
set _geometryRotated90(s) "w"
set _geometryRotated90(e) "s"
set _geometryRotated90(w) "n"

set _geometryRotated90(nw) "ne"
set _geometryRotated90(ne) "se"
set _geometryRotated90(sw) "nw"
set _geometryRotated90(se) "sw"

set _geometryRotated270(n) "w"
set _geometryRotated270(s) "e"
set _geometryRotated270(e) "n"
set _geometryRotated270(w) "s"

set _geometryRotated270(nw) "sw"
set _geometryRotated270(ne) "nw"
set _geometryRotated270(sw) "se"
set _geometryRotated270(se) "ne"

proc geomRotate {args} {
    global _geometryRotated90 _geometryRotated270

    getflag anti args
    behead direction args

    if { [geomIsCompass $direction] } {
	if {$anti} {
	    return $_geometryRotated270($direction)
	} else {
	    return $_geometryRotated90($direction)
	}
    } else {
	if {$anti} {
	    return [eval geomDirection \
		$_geometryRotated270([eval geomCompass $direction $args])]
	} else {
	    return [eval geomDirection \
		$_geometryRotated90([eval geomCompass $direction $args])]
	}
    }
}


## geomVector direction [handed]
#
# Translate a direction into a unit vector.
#
set _geometryVector(n) { 0 -1}
set _geometryVector(s) { 0  1}
set _geometryVector(e) { 1  0}
set _geometryVector(w) {-1  0}

set _oneOverRoot2 [expr 1.0 / sqrt(2.0)]

set _geometryVector(nw) [list -$_oneOverRoot2 -$_oneOverRoot2]
set _geometryVector(ne) [list  $_oneOverRoot2 -$_oneOverRoot2]
set _geometryVector(sw) [list -$_oneOverRoot2  $_oneOverRoot2]
set _geometryVector(se) [list  $_oneOverRoot2  $_oneOverRoot2]

proc geomVector {direction args} {
    global _geometryVector

    if { [geomIsCompass $direction] } {
	return $_geometryVector($direction)
    } else {
	return $_geometryVector([eval geomCompass $direction $args])
    }
}


## geomOffset direction amount
#
# Return a point representing a movement in the specified
# direction and by the specified amount.
#
# -- direction: Any valid direction. If a compound handedness
#    direction, this must be a distinct list:
#
#        geomOffset {top left} 10
# 
# -- amount: The distance of the offset.
#
proc geomOffset {direction amount} {
    return [pointScale [geomVector $direction] $amount]
}


## geomCompassNormalize compass
#
# Ensure that a compass direction is in the right order.
#
# --  compass: Any valid compass direction.
#
set _geometryCompassNormalize(wn) "nw"
set _geometryCompassNormalize(en) "ne"
set _geometryCompassNormalize(ws) "sw"
set _geometryCompassNormalize(es) "se"

proc geomCompassNormalize {compass} {
    global _geometryCompassNormalize

    if { [info exists _geometryCompassNormalize($compass)] } {
	return $_geometryCompassNormalize($compass)
    } else {
	return $compass
    }
}


## geomHandedNormalize direction [handed]
#
# Ensure that a handedness direction is in the right order.
#
# --  direction [handed]: Any valid handedness direction or
#     pair of directions.
#
proc geomHandedNormalize {direction args} {
    if { $args == "" } {
	return $direction
    } elseif { [geomVertical $direction] } {
	return [list $args $direction]
    } else {
	return [list $direction $args]
    }
}




##########################################################################
#
## Point functions
#
# These functions assist in dealing with co-ordinates expressed as
# ``points'' where a point is a pair (that is, a 2-list) of numbers.
# Note that most of these routines expect each point to be a distinct
# list, not concatenated into the argument list. For example,
#
#    pointAdd {1 2} {3 4}
#
# is correct, but
#
#    pointAdd 1 2 3 4
#
# is not.
#


## pointx, pointy
#
# Extract coordinates from a point
#
proc pointx {point} {
#    assert {[llength $point] == 2}

    return [lindex $point 0]
}

proc pointy {point} {
#    assert {[llength $point] == 2}

    return [lindex $point 1]
}

## pointIncrement point x y
#
# Add coordinates to a point.
#
proc pointIncrement {point x y} {
    return [list \
	    [expr [pointx $point] + $x] \
	    [expr [pointy $point] + $y] ]
}
  
## pointDecrement point x y
#
# Subtract coordinates from a point
#
proc pointDecrement {point x y} {
    return [list \
	    [expr [pointx $point] - $x] \
	    [expr [pointy $point] - $y] ]
}

## pointAdd point offset
#
# Add two points
#
proc pointAdd {point offset} {
    return [list \
	    [expr [lindex $point 0] + [lindex $offset 0]] \
	    [expr [lindex $point 1] + [lindex $offset 1]] ]
}


## pointSubtract point offset
#
# Subtract two points. That is, return _point_ - _offset_.
#
proc pointSubtract {point offset} {
    return [list \
	    [expr [pointx $point] - [pointx $offset]] \
	    [expr [pointy $point] - [pointy $offset]] ]
}
  

## pointScale point xscale {yscale}
#
# Scale a point. If only the first scale argument is given
# scale the point equally in both directions -- that is,
# given the point (x, y), return (x * xscale, y * xscale). If
# both scale arguments are given, scale the point by those amounts
# in the x and y directions.
#
proc pointScale {point xscale args} {
    if { $args == "" } {
	set yscale $xscale
    } else {
	set yscale [lindex $args 0]
    }

    return [list \
	    [expr [lindex $point 0] * $xscale] \
	    [expr [lindex $point 1] * $yscale] ]
}


## pointAxpy a x y
#
# Scale a point and then add it to another. In other words, the result
# is ax + y.
#
proc pointAxpy {a x y} {
    return [list \
	    [expr [lindex $x 0] * $a + [lindex $y 0]] \
	    [expr [lindex $x 1] * $a + [lindex $y 1]]]
}


## pointVxpy v x y
#
# Scale a point by a vector and then add it to another. In other
# words, the result is vx + y.
#
proc pointVxpy {v x y} {
    return [list \
	    [expr [lindex $x 0] * [lindex $v 0] + [lindex $y 0]] \
	    [expr [lindex $x 1] * [lindex $v 1] + [lindex $y 1]]]
}


## pointFlip point
#
# ``Flip'' a point about the origin. In other words, negate its
# two components.
#
proc pointFlip {point} {
    return [list \
	    [expr -[pointx $point]] \
	    [expr -[pointy $point]] ]
}


## pointMove point direction distance
#
# Move a point in the specified direction and
# by the specified amount
#
# -- point: The point to move.
#
# -- direction: Any valid direction. If a compound handedness
#    direction, this must be a distinct list:
#
#        pointMove {100 200} {top left} 10
# 
# -- distance: The distance to move.
#
proc pointMove {point direction distance} {
    return [pointAdd $point [geomOffset $direction $distance]]
}




##########################################################################
#
## Rectangle functions
#
# These functions assist in dealing with co-ordinates expressed as
# ``rectangles'' where a rectangle is a 4-list of numbers. As for
# points, rectangles are distinct lists, not concatenated into the
# argument list.
#

## rectx0, rectx1, recty0, rexty1
#
# Extract co-ordinates from a rectangle
#
proc rectx0 {rectangle} {
#    assert {[llength $rectangle] == 4}

    return [lindex $rectangle 0]
}

proc recty0 {rectangle} {
#    assert {[llength $rectangle] == 4}

    return [lindex $rectangle 1]
}

proc rectx1 {rectangle} {
#    assert {[llength $rectangle] == 4}

    return [lindex $rectangle 2]
}

proc recty1 {rectangle} {
#    assert {[llength $rectangle] == 4}

    return [lindex $rectangle 3]
}


## rectSize rect
#
# Return the size of a rectangle
#
proc rectSize {rect} {
    return [list \
	    [expr [rectx1 $rect] - [rectx0 $rect]] \
	    [expr [recty1 $rect] - [recty0 $rect]]]
}

## rectWidth rect
#
# Return the width of a rectangle
#
proc rectWidth {rect} {
    return [expr [rectx1 $rect] - [rectx0 $rect]]
}

## rectHeight rect
#
# Return the height of a rectangle
#
proc rectHeight {rect} {
    return [expr [recty1 $rect] - [recty0 $rect]]
}


## rectStrut rect
#
# Return the strut of a rectangle (as an interval). That is, return
# the two y-coordinates of the top and bottom edges of the rectangle.
#
proc rectStrut {rect} {
    return [list [recty1 $rect] [recty0 $rect]]
}

## rectSpan rect
#
# Return the span of a rectangle (as an interval). That is, return
# the two x-coordinates of the left and right edges of the rectangle.
#
proc rectSpan {rect} {
    return [list [rectx0 $rect] [rectx1 $rect]]
}

## rectStretch rect delta
#
# Stretch a rectangle on all sides. The rectangle will be larger
# on each side by 2 * _delta_ pixels.
#
# FIXME: There needs to be a way of stretching just the vertical
# or horizontal direction.
#
proc rectStretch {rect delta} {
    return [list \
	    [expr [rectx0 $rect] - $delta] \
	    [expr [recty0 $rect] - $delta] \
	    [expr [rectx1 $rect] + $delta] \
	    [expr [recty1 $rect] + $delta] ]
}

##
# rectShrink rect delta
#
# Shrink a rectangle on all sides. The rectangle will be smaller
# on each side by 2 * _delta_ pixels.
#
# FIXME: There needs to be a way of shrinking just the vertical
# or horizontal direction.
#
proc rectShrink {rect delta} {
    return [list \
	    [expr [rectx0 $rect] + $delta] \
	    [expr [recty0 $rect] + $delta] \
	    [expr [rectx1 $rect] - $delta] \
	    [expr [recty1 $rect] - $delta] ]
}

## rectAdd rect point
#
# Move a rectangle by the distance specified by the given point.
#
proc rectAdd {rect point} {
    return [list \
	    [expr [rectx0 $rect] + [pointx $point]] \
	    [expr [recty0 $rect] + [pointy $point]] \
	    [expr [rectx1 $rect] + [pointx $point]] \
	    [expr [recty1 $rect] + [pointy $point]] ]
}

## rectMove direction distance
#
# Move a rectangle by the specified distance in the  specified
# direction.
#
# -- rect: The rectangle to move.
#
# -- direction: Any valid direction. If a compound handedness
#    direction, this must be a distinct list:
#
#        rectMove {100 100 200 200} {bottom right} 10
# 
# -- distance: The distance to move.
#
proc rectMove {rect direction distance} {
    return [rectAdd $rect [geomOffset $direction $distance]]
}



##########################################################################
#
## Vertex functions
#
# The TK canvas support multi-edge lines, specified as a series of
# points. The vertex functions assist in dealing with this kind of
# line. Each of these functions accepts a ``line,'' which is a flat
# list of x and y co-ordinates, and an index, which specifies which
# ofits vertices to operate on.
#


## vertexAdd line index point
#
# Add a point to a vertex.
#
proc vertexAdd {line index point} {
    set m [expr $index*2]
    set n [expr $index*2+1]

    return [eval lreplace [list $line] $m $n \
	    [pointAdd [lrange $line $m $n] $point]]
}


## vertexMove line index direction distance
#
# Move a vertex in the specified direction and by the specified
# amount.
#
# -- line: The line containing the vertex.
#
# -- index: The index of the vertex.
#
# -- direction: Any valid direction. If a compound handedness
#    direction, this must be a distinct list:
#
#        vertexMove {100 100 200 200 300 100} 1 {bottom right} 10
# 
# -- distance: The distance to move.
#
proc vertexMove {line index direction distance} {
    return [vertexAdd $line $index [geomOffset $direction $distance]]
}


##########################################################################
#
## Oval functions
#
# The TK canvas supports ellipses, or ``ovals,'' as it calls them,
# but provides no facilities for finding points on the oval. Here are
# some routines to help out with some of this stuff.
#
# The equation of an ellipse with horizontal radius _a_ and
# vertical radius _b_ is
#
#        2        2
#       x        y
#      ----  +  ----  =  1
#        2        2
#       a        b
#
# Let m = y/x (the tangent of the angle to (x,y).
#
# Acknowledgement: these functions were derived from Wan-teh Chang's
# finite-state-machine editor code.
#

## pi and friends
#
set pi    3.1415926535
set twoPi [expr 2.0*$pi]
set piOverTwo [expr 0.5*$pi]
set threePiOverTwo [expr 1.5*$pi]


## ovalFromRect
#
# Given a rectangle specifying the boundary of the ellipse,
# return a description in the form {x y a b}.
#
proc ovalFromRect {rect} {
    assign x0 y0 x1 y1 $rect
    set a [expr ($x1 - $x0) / 2.0]
    set x [expr $x1 - $a]
    set b [expr ($y1 - $y0) / 2.0]
    set y [expr $y1 - $b]

    return [list $x $y $a $b]
}


## ovalBoundary
#
# Given an oval and an angle, return the point on the circumference
# of the oval. The angle _must_ be between -pi and 2*pi.
#
proc ovalBoundary {oval angle} {
    global pi twoPi piOverTwo threePiOverTwo

    assign x y a b $oval
    
    #
    # First calculate m, checking for vertical angles
    #
    if { $angle > $pi } {
	set angle [expr $angle - $twoPi]
    }

    if {       $angle > [expr $piOverTwo - 0.01] \
	    && $angle < [expr $piOverTwo + 0.01] } {
	return [list $x [expr $y - $b]]
    } elseif { $angle > [expr $threePiOverTwo - 0.01] \
	    && $angle < [expr $threePiOverTwo + 0.01] } {
	return [list $x [expr $y + $b]]
    }
    set m [expr tan($angle)]

    #
    # Calculate point based on m
    #
    return [ovalM $oval $m [expr $angle >= 0.0]]
}


## ovalPoint
#
# Given an oval and a point, return the point on the circumference
# of the oval which intersects the line from its center to the
# point.
#
proc ovalPoint {oval point} {
    assign x y a b $oval
    assign p q $point

    #
    # First calculate m, checking for vertical angles
    #
    if { $p == $x } {
	if { $q < $y } {
	    return [list $x [expr $y - $b]]
	} else {
	    return [list $x [expr $y + $b]]
	}
    }
    set m [expr ($y - $q) / ($p - $x)]

    #
    # Calculate point based on m
    #
    return [ovalM $oval $m [expr $q < $y]]
}


## ovalM
#
# Given an oval and the tangent from its center,
# return the point on the circumference of the oval.
# The positive argument specifies whether the angle
# is positive or negative.
#
proc ovalM {oval m positive} {
    assign x y a b $oval

    set offset [expr $a*$b / sqrt($b*$b + $m*$m*$a*$a)]
    set offset [expr ($m < 0) ^ ! $positive ? -$offset : $offset]

    set pointx [expr $x + $offset]
    set pointy [expr $y - $m * $offset]

    return [list $pointx $pointy]
}
